from migen import *
from migen.genlib.fifo import *
from migen.genlib.io import CRG

from migen.build.generic_platform import *
from migen.build.xilinx import XilinxPlatform
from migen.build.xilinx.programmer import XC3SProg


class FT601_245(Module):

    OP_READ = Constant(value=0, bits_sign=1)
    OP_WRITE = Constant(value=1, bits_sign=1)

    def __init__(self, io, width=32, depth=64):
        # rst = Signal()

        # Input FIFO for reads from FT600 into this FIFO
        incoming_fifo = self.incoming_fifo = SyncFIFO(width, depth)
        self.submodules += incoming_fifo

        # Output FIFO for write from this FIFO into FT600
        outgoing_fifo = self.outgoing_fifo = SyncFIFO(width, depth)
        self.submodules += outgoing_fifo

        # Shared control
        # Data bus is driven by both the FPGA and the FTDI. We use tristate.
        self.data_tri = data_tri = TSTriple(width)
        data_tri.oe.reset = 0
        assert io.data.nbits == width
        self.specials += data_tri.get_tristate(io.data)

        self.be_tri = be_tri = TSTriple(width//8)
        be_tri.oe.reset = 0
        assert io.be.nbits == width//8
        self.specials += be_tri.get_tristate(io.be)

        io.wr_n.reset = 1
        io.rd_n.reset = 1
        io.oe_n.reset = 1
        self.comb += io.wakeup_n.eq(0)

        rxf_n = self.rxf_n = Signal()
        txe_n = self.txe_n = Signal()
        self.comb += [
            rxf_n.eq(io.rxf_n),
            txe_n.eq(io.txe_n)
        ]

        wr_n = self.wr_n = Signal(reset=1)
        rd_n = self.rd_n = Signal(reset=1)
        oe_n = self.oe_n = Signal(reset=1)
        self.comb += [
            io.rd_n.eq(rd_n),
            io.wr_n.eq(wr_n),
            io.oe_n.eq(oe_n),
        ]

        curr_op = self.curr_op = Signal(reset=1)
        can_read = self.can_read = Signal()
        can_write = self.can_write = Signal()
        self.comb += can_read.eq(~rxf_n & incoming_fifo.writable)
        self.comb += can_write.eq(~txe_n & outgoing_fifo.readable)

        fsm = self.fsm = FSM(reset_state="IDLE")

        fsm.act("IDLE",
            If((can_read & can_write & (curr_op == self.OP_WRITE)) | (can_read & ~can_write),
                NextState("ASSERT_OE"),
               NextValue(curr_op, self.OP_READ),
            ).Elif((can_read & can_write & (curr_op == self.OP_READ)) | (~can_read & can_write),
                NextState("WRITE"),
                NextValue(curr_op, self.OP_WRITE),
            ).Else(
                NextState("IDLE"),
            )
        )

        fsm.act("ASSERT_OE",
            NextState("READ"),
        )

        fsm.act("READ",
            If(incoming_fifo.writable & ~rxf_n,
                NextState("READ"),
            ).Else(
                NextState("IDLE"),
            )
        )

        fsm.act("WRITE",
            If(outgoing_fifo.readable & ~txe_n,
                NextState("WRITE"),
            ).Else(
                NextState("IDLE"),
            )
        )

        self.submodules += fsm

        self.comb += [
            oe_n.eq(~(incoming_fifo.writable & ~rxf_n & (fsm.ongoing("ASSERT_OE") | fsm.ongoing("READ")))),
            rd_n.eq(~(incoming_fifo.writable & ~rxf_n & fsm.ongoing("READ"))),
            wr_n.eq(~(outgoing_fifo.readable & ~txe_n & fsm.ongoing("WRITE"))),
            data_tri.oe.eq(outgoing_fifo.readable & ~txe_n & fsm.ongoing("WRITE")),
            incoming_fifo.we.eq(incoming_fifo.writable & ~rxf_n & fsm.ongoing("READ")),
            outgoing_fifo.re.eq(outgoing_fifo.readable & ~txe_n & fsm.ongoing("WRITE"))
        ]

        self.comb += [
            incoming_fifo.din.eq(data_tri.i),
            data_tri.o.eq(outgoing_fifo.dout),
            be_tri.o.eq((width//8)**2-1),
            be_tri.oe.eq(data_tri.oe)
        ]


class FT601Top(Module):
    def __init__(self, platform):
        usb_fifo = platform.request("usb_fifo")

        crg = CRG(platform.request(platform.default_clk_name))
        self.submodules.crg = crg

        usb_phy = FT601_245(usb_fifo, depth=2048)
        self.submodules.usb_phy = usb_phy

        self.comb += [
            usb_phy.outgoing_fifo.din.eq(usb_phy.incoming_fifo.dout),
            usb_phy.outgoing_fifo.we.eq(usb_phy.incoming_fifo.readable & usb_phy.outgoing_fifo.writable),
            usb_phy.incoming_fifo.re.eq(usb_phy.incoming_fifo.readable & usb_phy.outgoing_fifo.writable),
        ]


_io = [
    ("clk100",  0, Pins("N21"), IOStandard("LVCMOS25")),

    ("usb_fifo", 0,
        Subsignal("data", Pins("K26 M26 K25 M25 L25 L23 L24 N26 M24 N24 M22 N23 M19 N22 N19 N18 P23 L22 P20 L20 P21 M20 T25 P24 R22 R21 T24 T22 P19 R17 R20 R16")),
        Subsignal("be", Pins("R23 P26 N17 R26")),
        #Subsignal("clk", Pins("Y23")),  # Used as main system clock clk100
        Subsignal("txe_n", Pins("T23")),
        Subsignal("rxf_n", Pins("R18")),
        Subsignal("wr_n", Pins("P25")),
        Subsignal("rd_n", Pins("R25")),
        Subsignal("oe_n", Pins("P15")),
        Subsignal("wakeup_n", Pins("N16")),
        IOStandard("LVCMOS25"),
    ),
]


class HSFPX002Platform(XilinxPlatform):
    default_clk_name = "clk100"
    default_clk_period = 10.000

    def __init__(self, programmer="xc3sprog"):
        self.programmer = programmer
        XilinxPlatform.__init__(self, "xc7a200tfbg676-2", _io, [], toolchain="vivado")

    def create_programmer(self):
        if self.programmer == "xc3sprog":
            return XC3SProg("jtaghs1")
        else:
            raise ValueError("{} programmer is not supported".format(self.programmer))


default_subtarget = FT601Top
default_platform = HSFPX002Platform

if __name__ == '__main__':
    p = HSFPX002Platform()
    m = FT601Top(p)
    # print(verilog.convert(m, ios=m.get_ios()))
    p.build(m, build_name="ft601_hsfpx002", build_dir="build_ft601_hsfpx002_top")